from __future__ import annotations

from typing import Sequence

import os

import numpy as np
from smac.runhistory.runhistory import RunHistory

from autosklearn.automl_common.common.utils.backend import Backend
from autosklearn.data.validation import SUPPORTED_FEAT_TYPES
from autosklearn.ensemble_building.run import Run
from autosklearn.ensembles.abstract_ensemble import AbstractEnsemble
from autosklearn.metrics import Scorer, calculate_losses
from autosklearn.pipeline.base import BasePipeline


class AbstractSingleModelEnsemble(AbstractEnsemble):
    """Ensemble consisting of a single model.

    Parameters
    ----------
    task_type: int
        An identifier indicating which task is being performed.

    metrics: Sequence[Scorer] | Scorer
        The metrics used to evaluate the models.

    backend : Backend
        Gives access to the backend of Auto-sklearn. Not used.

    random_state: int | RandomState | None = None
        Not used.
    """

    def __init__(
        self,
        task_type: int,
        metrics: Sequence[Scorer] | Scorer,
        backend: Backend,
        random_state: int | np.random.RandomState | None = None,
    ):
        self.weights_ = [1.0]
        self.task_type = task_type
        if isinstance(metrics, Sequence):
            self.metrics = metrics
        elif isinstance(metrics, Scorer):
            self.metrics = [metrics]
        else:
            raise TypeError(type(metrics))
        self.random_state = random_state
        self.backend = backend

    def fit(
        self,
        base_models_predictions: np.ndarray | list[np.ndarray],
        true_targets: np.ndarray,
        model_identifiers: list[tuple[int, int, float]],
        runs: Sequence[Run],
        X_data: SUPPORTED_FEAT_TYPES | None = None,
    ) -> AbstractSingleModelEnsemble:
        """Fit the ensemble

        Parameters
        ----------
        base_models_predictions: np.ndarray
            shape = (n_base_models, n_data_points, n_targets)
            n_targets is the number of classes in case of classification,
            n_targets is 0 or 1 in case of regression

            Can be a list of 2d numpy arrays as well to prevent copying all
            predictions into a single, large numpy array.

        true_targets : array of shape [n_targets]

        model_identifiers : identifier for each base model.
            Can be used for practical text output of the ensemble.

        runs: Sequence[Run]
            Additional information for each run executed by SMAC that was
            considered by the ensemble builder.

        X_data : list-like | sparse matrix | None = None

        Returns
        -------
        self
        """
        return self

    def predict(self, predictions: np.ndarray | list[np.ndarray]) -> np.ndarray:
        """Select the predictions of the selected model.

        Parameters
        ----------
        base_models_predictions : np.ndarray
            shape = (n_base_models, n_data_points, n_targets)
            Same as in the fit method.

        Returns
        -------
        np.ndarray
        """
        return predictions[0]

    def __str__(self) -> str:
        return "%s:\n\tMembers: %s" "\n\tWeights: %s\n\tIdentifiers: [%s]" % (
            self.__class__.__name__,
            self.indices_,  # type: ignore [attr-defined]
            self.weights_,
            self.identifiers_[0],  # type: ignore [attr-defined]
        )

    def get_models_with_weights(
        self, models: dict[tuple[int, int, float], BasePipeline]
    ) -> list[tuple[float, BasePipeline]]:
        """List of (weight, model) pairs for the model selected by this ensemble.

        Parameters
        ----------
        models : dict {identifier : model object}
            The identifiers are the same as the one presented to the fit()
            method. Models can be used for nice printing.

        Returns
        -------
        list[tuple[float, BasePipeline]]
        """
        return [(self.weights_[0], models[self.identifiers_[0]])]  # type: ignore [attr-defined]  # noqa: E501

    def get_identifiers_with_weights(
        self,
    ) -> list[tuple[tuple[int, int, float], float]]:
        """Return a (identifier, weight)-pairs for the model selected by this ensemble.

        Parameters
        ----------
        models : dict {identifier : model object}
            The identifiers are the same as the one presented to the fit()
            method. Models can be used for nice printing.

        Returns
        -------
        list[tuple[tuple[int, int, float], float]
        """
        return list(zip(self.identifiers_, self.weights_))  # type: ignore [attr-defined]  # noqa: E501

    def get_selected_model_identifiers(self) -> list[tuple[int, int, float]]:
        """Return identifier of models in the ensemble.

        This includes models which have a weight of zero!

        Returns
        -------
        list
        """
        return self.identifiers_  # type: ignore [attr-defined]

    def get_validation_performance(self) -> float:
        """Return validation performance of ensemble.

        In case of multi-objective problem, only the first metric will be returned.

        Return
        ------
        float
        """
        return self.best_model_score_  # type: ignore [attr-defined]


class SingleModelEnsemble(AbstractSingleModelEnsemble):
    """Ensemble consisting of a single model.

    This class is used by the :class:`MultiObjectiveDummyEnsemble` to represent
    ensembles consisting of a single model, and this class should not be used
    on its own.

    Do not use by yourself!

    Parameters
    ----------
    task_type: int
        An identifier indicating which task is being performed.

    metrics: Sequence[Scorer] | Scorer
        The metrics used to evaluate the models.

    backend : Backend
        Gives access to the backend of Auto-sklearn. Not used.

    model_index : int
        Index of the model that constitutes the ensemble. This index will
        be used to select the correct predictions that will be passed during
        ``fit`` and ``predict``.

    random_state: int | RandomState | None = None
        Not used.
    """

    def __init__(
        self,
        task_type: int,
        metrics: Sequence[Scorer] | Scorer,
        backend: Backend,
        model_index: int,
        random_state: int | np.random.RandomState | None = None,
    ):
        super().__init__(
            task_type=task_type,
            metrics=metrics,
            random_state=random_state,
            backend=backend,
        )
        self.indices_ = [model_index]

    def fit(
        self,
        base_models_predictions: np.ndarray | list[np.ndarray],
        true_targets: np.ndarray,
        model_identifiers: list[tuple[int, int, float]],
        runs: Sequence[Run],
        X_data: SUPPORTED_FEAT_TYPES | None = None,
    ) -> SingleModelEnsemble:
        """Dummy implementation of the ``fit`` method.

        Actualy work of passing the model index is done in the constructor. This
        method only stores the identifier of the selected model and computes it's
        validation loss.

        Parameters
        ----------
        base_models_predictions: np.ndarray
            shape = (n_base_models, n_data_points, n_targets)
            n_targets is the number of classes in case of classification,
            n_targets is 0 or 1 in case of regression

            Can be a list of 2d numpy arrays as well to prevent copying all
            predictions into a single, large numpy array.

        true_targets : array of shape [n_targets]

        model_identifiers : identifier for each base model.
            Can be used for practical text output of the ensemble.

        runs: Sequence[Run]
            Additional information for each run executed by SMAC that was
            considered by the ensemble builder. Not used.

        X_data : list-like | spmatrix | None = None
           X data to feed to a metric if it requires it

        Returns
        -------
        self
        """
        self.identifiers_ = [model_identifiers[self.indices_[0]]]
        loss = calculate_losses(
            solution=true_targets,
            prediction=base_models_predictions[self.indices_[0]],
            task_type=self.task_type,
            metrics=self.metrics,
            X_data=X_data,
        )
        self.best_model_score_ = loss[self.metrics[0].name]
        return self


class SingleBest(AbstractSingleModelEnsemble):
    """Ensemble consisting of the single best model.

    Parameters
    ----------
    task_type: int
        An identifier indicating which task is being performed.

    metrics: Sequence[Scorer] | Scorer
        The metrics used to evaluate the models.

    random_state: int | RandomState | None = None
        Not used.

    backend : Backend
        Gives access to the backend of Auto-sklearn. Not used.
    """

    def __init__(
        self,
        task_type: int,
        metrics: Sequence[Scorer] | Scorer,
        backend: Backend,
        random_state: int | np.random.RandomState | None = None,
    ):
        super().__init__(
            task_type=task_type,
            metrics=metrics,
            random_state=random_state,
            backend=backend,
        )

    def fit(
        self,
        base_models_predictions: np.ndarray | list[np.ndarray],
        true_targets: np.ndarray,
        model_identifiers: list[tuple[int, int, float]],
        runs: Sequence[Run],
        X_data: SUPPORTED_FEAT_TYPES | None = None,
    ) -> SingleBest:
        """Select the single best model.

        Parameters
        ----------
        base_models_predictions: np.ndarray
            shape = (n_base_models, n_data_points, n_targets)
            n_targets is the number of classes in case of classification,
            n_targets is 0 or 1 in case of regression

            Can be a list of 2d numpy arrays as well to prevent copying all
            predictions into a single, large numpy array.

        true_targets : array of shape [n_targets]

        model_identifiers : identifier for each base model.
            Can be used for practical text output of the ensemble.

        runs: Sequence[Run]
            Additional information for each run executed by SMAC that was
            considered by the ensemble builder. Not used.

        X_data : array-like | sparse matrix | None = None

        Returns
        -------
         self
        """
        losses = [
            calculate_losses(
                solution=true_targets,
                prediction=base_model_prediction,
                task_type=self.task_type,
                metrics=self.metrics,
                X_data=X_data,
            )[self.metrics[0].name]
            for base_model_prediction in base_models_predictions
        ]
        argmin = np.argmin(losses)
        self.indices_ = [argmin]
        self.identifiers_ = [model_identifiers[argmin]]
        self.best_model_score_ = losses[argmin]
        return self


class SingleBestFromRunhistory(AbstractSingleModelEnsemble):
    """
    In the case of a crash, this class searches
    for the best individual model.

    Such model is returned as an ensemble of a single
    object, to comply with the expected interface of an
    AbstractEnsemble.

    Do not use by yourself!
    """

    def __init__(
        self,
        task_type: int,
        metrics: Sequence[Scorer] | Scorer,
        backend: Backend,
        run_history: RunHistory,
        seed: int,
        random_state: int | np.random.RandomState | None = None,
    ):
        super().__init__(
            task_type=task_type,
            metrics=metrics,
            random_state=random_state,
            backend=backend,
        )

        # The seed here is seperate from RandomState and is used to indiicate a
        # directory for the backend to search in
        self.seed = seed
        self.indices_ = [0]
        self.weights_ = [1.0]
        self.run_history = run_history
        self.identifiers_ = self.get_identifiers_from_run_history()

    def get_identifiers_from_run_history(self) -> list[tuple[int, int, float]]:
        """Parses the run history, to identify the best performing model

        Populates the identifiers attribute, which is used by the backend to access
        the actual model.
        """
        best_model_identifier = []
        best_model_score = self.metrics[0]._worst_possible_result

        for run_key in self.run_history.data.keys():
            run_value = self.run_history.data[run_key]
            print(run_key, run_value)
            if len(self.metrics) == 1:
                cost = run_value.cost
            else:
                cost = run_value.cost[0]
            score = self.metrics[0]._optimum - (self.metrics[0]._sign * cost)

            if (score > best_model_score and self.metrics[0]._sign > 0) or (
                score < best_model_score and self.metrics[0]._sign < 0
            ):

                # Make sure that the individual best model actually exists
                model_dir = self.backend.get_numrun_directory(
                    self.seed,
                    run_value.additional_info["num_run"],
                    run_key.budget,
                )
                model_file_name = self.backend.get_model_filename(
                    self.seed,
                    run_value.additional_info["num_run"],
                    run_key.budget,
                )
                file_path = os.path.join(model_dir, model_file_name)
                if not os.path.exists(file_path):
                    continue

                best_model_identifier = [
                    (
                        self.seed,
                        run_value.additional_info["num_run"],
                        run_key.budget,
                    )
                ]
                best_model_score = score

        if not best_model_identifier:
            raise ValueError(
                "No valid model found in run history. This means smac was not able to"
                " fit a valid model. Please check the log file for errors."
            )

        self.best_model_score_ = best_model_score

        return best_model_identifier



from typing import Optional

from functools import partial

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    Constant,
    UniformFloatHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.components.feature_preprocessing.select_percentile import (
    SelectPercentileBase,
)
from autosklearn.pipeline.constants import (
    DENSE,
    INPUT,
    SIGNED_DATA,
    SPARSE,
    UNSIGNED_DATA,
)


class SelectPercentileClassification(
    SelectPercentileBase, AutoSklearnPreprocessingAlgorithm
):
    def __init__(self, percentile, score_func="chi2", random_state=None):
        """Parameters:
        random state : ignored

        score_func : callable, Function taking two arrays X and y, and
                     returning a pair of arrays (scores, pvalues).
        """
        import sklearn.feature_selection

        self.random_state = random_state  # We don't use this
        self.percentile = int(float(percentile))
        if score_func == "chi2":
            self.score_func = sklearn.feature_selection.chi2
        elif score_func == "f_classif":
            self.score_func = sklearn.feature_selection.f_classif
        elif score_func == "mutual_info":
            self.score_func = partial(
                sklearn.feature_selection.mutual_info_classif,
                random_state=self.random_state,
            )
        else:
            raise ValueError(
                "score_func must be in ('chi2, 'f_classif', 'mutual_info'), "
                "but is: %s" % score_func
            )

    def fit(self, X, y):
        import scipy.sparse
        import sklearn.feature_selection

        self.preprocessor = sklearn.feature_selection.SelectPercentile(
            score_func=self.score_func,
            percentile=self.percentile,
        )

        # Because the pipeline guarantees that each feature is positive,
        # clip all values below zero to zero
        if self.score_func == sklearn.feature_selection.chi2:
            if scipy.sparse.issparse(X):
                X.data[X.data < 0] = 0.0
            else:
                X[X < 0] = 0.0

        self.preprocessor.fit(X, y)
        return self

    def transform(self, X):
        import scipy.sparse
        import sklearn.feature_selection

        # Because the pipeline guarantees that each feature is positive,
        # clip all values below zero to zero
        if self.score_func == sklearn.feature_selection.chi2:
            if scipy.sparse.issparse(X):
                X.data[X.data < 0] = 0.0
            else:
                X[X < 0] = 0.0

        if self.preprocessor is None:
            raise NotImplementedError()
        Xt = self.preprocessor.transform(X)
        if Xt.shape[1] == 0:
            raise ValueError("%s removed all features." % self.__class__.__name__)
        return Xt

    @staticmethod
    def get_properties(dataset_properties=None):
        data_type = UNSIGNED_DATA
        if dataset_properties is not None:
            signed = dataset_properties.get("signed")
            if signed is not None:
                data_type = SIGNED_DATA if signed is True else UNSIGNED_DATA

        return {
            "shortname": "SPC",
            "name": "Select Percentile Classification",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": False,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (SPARSE, DENSE, data_type),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        percentile = UniformFloatHyperparameter(
            name="percentile", lower=1, upper=99, default_value=50
        )

        score_func = CategoricalHyperparameter(
            name="score_func",
            choices=["chi2", "f_classif", "mutual_info"],
            default_value="chi2",
        )
        if dataset_properties is not None:
            # Chi2 can handle sparse data, so we respect this
            if "sparse" in dataset_properties and dataset_properties["sparse"]:
                score_func = Constant(name="score_func", value="chi2")

        cs = ConfigurationSpace()
        cs.add_hyperparameters([percentile, score_func])

        return cs



from typing import Optional

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformIntegerHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnRegressionAlgorithm
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, SPARSE, UNSIGNED_DATA


class KNearestNeighborsRegressor(AutoSklearnRegressionAlgorithm):
    def __init__(self, n_neighbors, weights, p, random_state=None):
        self.n_neighbors = n_neighbors
        self.weights = weights
        self.p = p
        self.random_state = random_state

    def fit(self, X, y):
        import sklearn.neighbors

        self.n_neighbors = int(self.n_neighbors)
        self.p = int(self.p)

        self.estimator = sklearn.neighbors.KNeighborsRegressor(
            n_neighbors=self.n_neighbors, weights=self.weights, p=self.p
        )

        if y.ndim == 2 and y.shape[1] == 1:
            y = y.flatten()

        self.estimator.fit(X, y)
        return self

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError()
        return self.estimator.predict(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "KNN",
            "name": "K-Nearest Neighbor Classification",
            "handles_regression": True,
            "handles_classification": False,
            "handles_multiclass": False,
            "handles_multilabel": False,
            "handles_multioutput": True,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        cs = ConfigurationSpace()

        n_neighbors = UniformIntegerHyperparameter(
            name="n_neighbors", lower=1, upper=100, log=True, default_value=1
        )
        weights = CategoricalHyperparameter(
            name="weights", choices=["uniform", "distance"], default_value="uniform"
        )
        p = CategoricalHyperparameter(name="p", choices=[1, 2], default_value=2)

        cs.add_hyperparameters([n_neighbors, weights, p])

        return cs



from typing import Dict, Optional, Tuple, Union

from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.components.data_preprocessing.rescaling.abstract_rescaling import (  # noqa: E501
    Rescaling,
)
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class NoRescalingComponent(Rescaling, AutoSklearnPreprocessingAlgorithm):
    def fit(
        self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
    ) -> "AutoSklearnPreprocessingAlgorithm":
        self.preprocessor = "passthrough"
        return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        return X

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "NoRescaling",
            "name": "NoRescaling",
            "handles_missing_values": False,
            "handles_nominal_values": False,
            "handles_numerical_features": True,
            "prefers_data_scaled": False,
            "prefers_data_normalized": False,
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            # TODO find out if this is right!
            "handles_sparse": True,
            "handles_dense": True,
            "input": (SPARSE, DENSE, UNSIGNED_DATA),
            "output": (INPUT,),
            "preferred_dtype": None,
        }



from typing import Optional

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
    UnParametrizedHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import (
    AutoSklearnClassificationAlgorithm,
    IterativeComponentWithSampleWeight,
)
from autosklearn.pipeline.constants import DENSE, PREDICTIONS, SPARSE, UNSIGNED_DATA
from autosklearn.pipeline.implementations.util import (
    convert_multioutput_multiclass_to_multilabel,
)
from autosklearn.util.common import check_for_bool, check_none


class RandomForest(
    IterativeComponentWithSampleWeight,
    AutoSklearnClassificationAlgorithm,
):
    def __init__(
        self,
        criterion,
        max_features,
        max_depth,
        min_samples_split,
        min_samples_leaf,
        min_weight_fraction_leaf,
        bootstrap,
        max_leaf_nodes,
        min_impurity_decrease,
        random_state=None,
        n_jobs=1,
        class_weight=None,
    ):
        self.n_estimators = self.get_max_iter()
        self.criterion = criterion
        self.max_features = max_features
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.bootstrap = bootstrap
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.random_state = random_state
        self.n_jobs = n_jobs
        self.class_weight = class_weight
        self.estimator = None

    @staticmethod
    def get_max_iter():
        return 512

    def get_current_iter(self):
        return self.estimator.n_estimators

    def iterative_fit(self, X, y, sample_weight=None, n_iter=1, refit=False):
        from sklearn.ensemble import RandomForestClassifier

        if refit:
            self.estimator = None

        if self.estimator is None:
            self.n_estimators = int(self.n_estimators)
            if check_none(self.max_depth):
                self.max_depth = None
            else:
                self.max_depth = int(self.max_depth)

            self.min_samples_split = int(self.min_samples_split)
            self.min_samples_leaf = int(self.min_samples_leaf)
            self.min_weight_fraction_leaf = float(self.min_weight_fraction_leaf)

            if self.max_features not in ("sqrt", "log2", "auto"):
                max_features = int(X.shape[1] ** float(self.max_features))
            else:
                max_features = self.max_features

            self.bootstrap = check_for_bool(self.bootstrap)

            if check_none(self.max_leaf_nodes):
                self.max_leaf_nodes = None
            else:
                self.max_leaf_nodes = int(self.max_leaf_nodes)

            self.min_impurity_decrease = float(self.min_impurity_decrease)

            # initial fit of only increment trees
            self.estimator = RandomForestClassifier(
                n_estimators=n_iter,
                criterion=self.criterion,
                max_features=max_features,
                max_depth=self.max_depth,
                min_samples_split=self.min_samples_split,
                min_samples_leaf=self.min_samples_leaf,
                min_weight_fraction_leaf=self.min_weight_fraction_leaf,
                bootstrap=self.bootstrap,
                max_leaf_nodes=self.max_leaf_nodes,
                min_impurity_decrease=self.min_impurity_decrease,
                random_state=self.random_state,
                n_jobs=self.n_jobs,
                class_weight=self.class_weight,
                warm_start=True,
            )
        else:
            self.estimator.n_estimators += n_iter
            self.estimator.n_estimators = min(
                self.estimator.n_estimators, self.n_estimators
            )

        self.estimator.fit(X, y, sample_weight=sample_weight)
        return self

    def configuration_fully_fitted(self):
        if self.estimator is None:
            return False

        return not len(self.estimator.estimators_) < self.n_estimators

    def predict(self, X):
        if self.estimator is None:
            raise NotImplementedError
        return self.estimator.predict(X)

    def predict_proba(self, X):
        if self.estimator is None:
            raise NotImplementedError()
        probas = self.estimator.predict_proba(X)
        probas = convert_multioutput_multiclass_to_multilabel(probas)
        return probas

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "RF",
            "name": "Random Forest Classifier",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        cs = ConfigurationSpace()
        criterion = CategoricalHyperparameter(
            "criterion", ["gini", "entropy"], default_value="gini"
        )

        # The maximum number of features used in the forest is calculated as
        # m^max_features, where m is the total number of features, and max_features
        # is the hyperparameter specified below. The default is 0.5, which yields
        # sqrt(m) features as max_features in the estimator.
        # This corresponds with Geurts' heuristic.
        max_features = UniformFloatHyperparameter(
            "max_features", 0.0, 1.0, default_value=0.5
        )

        max_depth = UnParametrizedHyperparameter("max_depth", "None")
        min_samples_split = UniformIntegerHyperparameter(
            "min_samples_split", 2, 20, default_value=2
        )
        min_samples_leaf = UniformIntegerHyperparameter(
            "min_samples_leaf", 1, 20, default_value=1
        )
        min_weight_fraction_leaf = UnParametrizedHyperparameter(
            "min_weight_fraction_leaf", 0.0
        )
        max_leaf_nodes = UnParametrizedHyperparameter("max_leaf_nodes", "None")
        min_impurity_decrease = UnParametrizedHyperparameter(
            "min_impurity_decrease", 0.0
        )
        bootstrap = CategoricalHyperparameter(
            "bootstrap", ["True", "False"], default_value="True"
        )
        cs.add_hyperparameters(
            [
                criterion,
                max_features,
                max_depth,
                min_samples_split,
                min_samples_leaf,
                min_weight_fraction_leaf,
                max_leaf_nodes,
                bootstrap,
                min_impurity_decrease,
            ]
        )
        return cs



from abc import ABCMeta
from typing import Dict, Optional, Union

import numpy as np
import scipy.sparse
from ConfigSpace import Configuration
from sklearn.pipeline import Pipeline

import autosklearn.pipeline.create_searchspace_util
from autosklearn.askl_typing import FEAT_TYPE_TYPE

from .components.base import AutoSklearnChoice, AutoSklearnComponent

DATASET_PROPERTIES_TYPE = Dict[str, Union[str, int, bool]]
PIPELINE_DATA_DTYPE = Union[
    np.ndarray,
    scipy.sparse.bsr_matrix,
    scipy.sparse.coo_matrix,
    scipy.sparse.csc_matrix,
    scipy.sparse.csr_matrix,
    scipy.sparse.dia_matrix,
    scipy.sparse.dok_matrix,
    scipy.sparse.lil_matrix,
]


class BasePipeline(Pipeline):
    """Base class for all pipeline objects.

    Notes
    -----
    This class should not be instantiated, only subclassed."""

    __metaclass__ = ABCMeta

    def __init__(
        self,
        config=None,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        steps=None,
        dataset_properties=None,
        include=None,
        exclude=None,
        random_state=None,
        init_params=None,
    ):

        self.init_params = init_params if init_params is not None else {}
        self.include = include if include is not None else {}
        self.exclude = exclude if exclude is not None else {}
        self.dataset_properties = (
            dataset_properties if dataset_properties is not None else {}
        )
        self.random_state = random_state
        self.feat_type = feat_type

        if steps is None:
            self.steps = self._get_pipeline_steps(
                feat_type=feat_type, dataset_properties=dataset_properties
            )
        else:
            self.steps = steps

        self._validate_include_exclude_params()

        self.config_space = self.get_hyperparameter_search_space(feat_type=feat_type)

        if config is None:
            self.config = self.config_space.get_default_configuration()
        else:
            if isinstance(config, dict):
                config = Configuration(self.config_space, config)
            if self.config_space != config.configuration_space:
                print(self.config_space._children)
                print(config.configuration_space._children)
                import difflib

                diff = difflib.unified_diff(
                    str(self.config_space).splitlines(),
                    str(config.configuration_space).splitlines(),
                )
                diff = "\n".join(diff)
                raise ValueError(
                    "Configuration passed does not come from the "
                    "same configuration space. Differences are: "
                    "%s" % diff
                )
            self.config = config

        self.set_hyperparameters(
            self.config, feat_type=feat_type, init_params=init_params
        )

        super().__init__(steps=self.steps)

        self._additional_run_info = {}

    def fit(self, X, y, **fit_params):
        """Fit the selected algorithm to the training data.

        Parameters
        ----------
        X : array-like or sparse, shape = (n_samples, n_features)
            Training data. The preferred type of the matrix (dense or sparse)
            depends on the estimator selected.

        y : array-like
            Targets

        fit_params : dict
            See the documentation of sklearn.pipeline.Pipeline for formatting
            instructions.

        Returns
        -------
        self : returns an instance of self.

        Raises
        ------
        NoModelException
            NoModelException is raised if fit() is called without specifying
            a classification algorithm first.
        """
        X, fit_params = self.fit_transformer(X, y, **fit_params)
        self.fit_estimator(X, y, **fit_params)
        return self

    def fit_transformer(self, X, y, fit_params=None):
        self.num_targets = 1 if len(y.shape) == 1 else y.shape[1]
        if fit_params is None:
            fit_params = {}
        fit_params = {
            key.replace(":", "__"): value for key, value in fit_params.items()
        }
        fit_params_steps = self._check_fit_params(**fit_params)
        Xt = self._fit(X, y, **fit_params_steps)
        return Xt, fit_params_steps[self.steps[-1][0]]

    def fit_estimator(self, X, y, **fit_params):
        fit_params = {
            key.replace(":", "__"): value for key, value in fit_params.items()
        }
        self._final_estimator.fit(X, y, **fit_params)
        return self

    def iterative_fit(self, X, y, n_iter=1, **fit_params):
        self._final_estimator.iterative_fit(X, y, n_iter=n_iter, **fit_params)

    def estimator_supports_iterative_fit(self):
        return self._final_estimator.estimator_supports_iterative_fit()

    def get_max_iter(self):
        if self.estimator_supports_iterative_fit():
            return self._final_estimator.get_max_iter()
        else:
            raise NotImplementedError()

    def configuration_fully_fitted(self):
        return self._final_estimator.configuration_fully_fitted()

    def get_current_iter(self):
        return self._final_estimator.get_current_iter()

    def predict(self, X, batch_size=None):
        """Predict the classes using the selected model.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)

        batch_size: int or None, defaults to None
            batch_size controls whether the pipeline will be
            called on small chunks of the data. Useful when calling the
            predict method on the whole array X results in a MemoryError.

        Returns
        -------
        array, shape=(n_samples,) if n_classes == 2 else (n_samples, n_classes)
            Returns the predicted values"""

        if batch_size is None:
            return super().predict(X).astype(self._output_dtype)
        else:
            if not isinstance(batch_size, int):
                raise ValueError(
                    "Argument 'batch_size' must be of type int, "
                    "but is '%s'" % type(batch_size)
                )
            if batch_size <= 0:
                raise ValueError(
                    "Argument 'batch_size' must be positive, " "but is %d" % batch_size
                )

            else:
                if self.num_targets == 1:
                    y = np.zeros((X.shape[0],), dtype=self._output_dtype)
                else:
                    y = np.zeros(
                        (X.shape[0], self.num_targets), dtype=self._output_dtype
                    )

                # Copied and adapted from the scikit-learn GP code
                for k in range(max(1, int(np.ceil(float(X.shape[0]) / batch_size)))):
                    batch_from = k * batch_size
                    batch_to = min([(k + 1) * batch_size, X.shape[0]])
                    y[batch_from:batch_to] = self.predict(
                        X[batch_from:batch_to], batch_size=None
                    )

                return y

    def set_hyperparameters(
        self,
        configuration,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        init_params=None,
    ):
        self.config = configuration

        for node_idx, n_ in enumerate(self.steps):
            node_name, node = n_

            sub_configuration_space = node.get_hyperparameter_search_space(
                feat_type=feat_type, dataset_properties=self.dataset_properties
            )
            sub_config_dict = {}
            for param in configuration:
                if param.startswith("%s:" % node_name):
                    value = configuration[param]
                    new_name = param.replace("%s:" % node_name, "", 1)
                    sub_config_dict[new_name] = value

            sub_configuration = Configuration(
                sub_configuration_space, values=sub_config_dict
            )

            if init_params is not None:
                sub_init_params_dict = {}
                for param in init_params:
                    if param.startswith("%s:" % node_name):
                        value = init_params[param]
                        new_name = param.replace("%s:" % node_name, "", 1)
                        sub_init_params_dict[new_name] = value
            else:
                sub_init_params_dict = None

            if isinstance(
                node, (AutoSklearnChoice, AutoSklearnComponent, BasePipeline)
            ):
                node.set_hyperparameters(
                    feat_type=feat_type,
                    configuration=sub_configuration,
                    init_params=sub_init_params_dict,
                )
            else:
                raise NotImplementedError("Not supported yet!")

        # In-code check to make sure init params
        # is checked after pipeline creation
        self._check_init_params_honored(init_params)

        return self

    def get_hyperparameter_search_space(
        self, feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        """Return the configuration space for the CASH problem.

        Returns
        -------
        cs : ConfigSpace.configuration_space.Configuration
            The configuration space describing the AutoSklearnClassifier.

        """
        if not hasattr(self, "config_space") or self.config_space is None:
            self.config_space = self._get_hyperparameter_search_space(
                feat_type=feat_type,
                include=self.include,
                exclude=self.exclude,
                dataset_properties=self.dataset_properties,
            )
        return self.config_space

    def _get_hyperparameter_search_space(
        self,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        include=None,
        exclude=None,
        dataset_properties=None,
    ):
        """Return the configuration space for the CASH problem.

        This method should be called by the method
        get_hyperparameter_search_space of a subclass. After the subclass
        assembles a list of available estimators and preprocessor components,
        _get_hyperparameter_search_space can be called to do the work of
        creating the actual
        ConfigSpace.configuration_space.ConfigurationSpace object.

        Parameters
        ----------
        feat_type: dict
            python dictionary which maps the columns of the dataset to the data types

        estimator_name : str
            Name of the estimator hyperparameter which will be used in the
            configuration space. For a classification task, this would be
            'classifier'.

        estimator_components : dict {name: component}
            Dictionary with all estimator components to be included in the
            configuration space.

        preprocessor_components : dict {name: component}
            Dictionary with all preprocessor components to be included in the
            configuration space. .

        always_active : list of str
            A list of components which will always be active in the pipeline.
            This is useful for components like imputation which have
            hyperparameters to be configured, but which do not have any parent.

        default_estimator : str
            Default value for the estimator hyperparameter.

        Returns
        -------
        cs : ConfigSpace.configuration_space.Configuration
            The configuration space describing the AutoSklearnClassifier.
        """
        raise NotImplementedError()

    def _get_base_search_space(
        self,
        cs,
        dataset_properties,
        include,
        exclude,
        pipeline,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
    ):
        if include is None:
            if self.include is None:
                include = {}
            else:
                include = self.include

        keys = [pair[0] for pair in pipeline]
        for key in include:
            if key not in keys:
                raise ValueError(
                    "Invalid key in include: %s; should be one " "of %s" % (key, keys)
                )

        if exclude is None:
            if self.exclude is None:
                exclude = {}
            else:
                exclude = self.exclude

        keys = [pair[0] for pair in pipeline]
        for key in exclude:
            if key not in keys:
                raise ValueError(
                    "Invalid key in exclude: %s; should be one " "of %s" % (key, keys)
                )

        if "sparse" not in dataset_properties:
            # This dataset is probably dense
            dataset_properties["sparse"] = False
        if "signed" not in dataset_properties:
            # This dataset probably contains unsigned data
            dataset_properties["signed"] = False

        matches = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=pipeline,
            dataset_properties=dataset_properties,
            include=include,
            exclude=exclude,
        )

        # Now we have only legal combinations at this step of the pipeline
        # Simple sanity checks
        assert np.sum(matches) != 0, "No valid pipeline found."

        assert np.sum(matches) <= np.size(
            matches
        ), "'matches' is not binary; %s <= %d, %s" % (
            str(np.sum(matches)),
            np.size(matches),
            str(matches.shape),
        )

        # Iterate each dimension of the matches array (each step of the
        # pipeline) to see if we can add a hyperparameter for that step
        for node_idx, n_ in enumerate(pipeline):
            node_name, node = n_

            is_choice = isinstance(node, AutoSklearnChoice)

            # if the node isn't a choice we can add it immediately because it
            #  must be active (if it wasn't, np.sum(matches) would be zero
            if not is_choice:
                cs.add_configuration_space(
                    node_name,
                    node.get_hyperparameter_search_space(
                        dataset_properties=dataset_properties, feat_type=feat_type
                    ),
                )
            # If the node is a choice, we have to figure out which of its
            #  choices are actually legal choices
            else:
                choices_list = (
                    autosklearn.pipeline.create_searchspace_util.find_active_choices(
                        matches,
                        node,
                        node_idx,
                        dataset_properties,
                        include.get(node_name),
                        exclude.get(node_name),
                    )
                )
                sub_config_space = node.get_hyperparameter_search_space(
                    feat_type=feat_type,
                    dataset_properties=dataset_properties,
                    include=choices_list,
                )
                cs.add_configuration_space(node_name, sub_config_space)

        # And now add forbidden parameter configurations
        # According to matches
        if np.sum(matches) < np.size(matches):
            cs = autosklearn.pipeline.create_searchspace_util.add_forbidden(
                conf_space=cs,
                pipeline=pipeline,
                matches=matches,
                dataset_properties=dataset_properties,
                include=include,
                exclude=exclude,
            )

        return cs

    def _check_init_params_honored(self, init_params):
        """
        Makes sure that init params is honored at the implementation level
        """
        if init_params is None or len(init_params) < 1:
            # None/empty dict, so no further check required
            return

        # There is the scenario, where instance is passed as an argument to the
        # init_params 'instance': '{"task_id": "73543c4a360aa24498c0967fbc2f926b"}'}
        # coming from smac instance. Remove this key to make the testing stricter
        init_params.pop("instance", None)

        for key, value in init_params.items():

            if ":" not in key:
                raise ValueError(
                    "Unsupported argument to init_params {}."
                    "When using init_params, a hierarchical format like "
                    "node_name:parameter must be provided.".format(key)
                )
            node_name = key.split(":", 1)[0]
            if node_name not in self.named_steps.keys():
                raise ValueError(
                    "The current node name specified via key={} of init_params "
                    "is not valid. Valid node names are {}".format(
                        key, self.named_steps.keys()
                    )
                )
                continue
            variable_name = key.split(":")[-1]
            node = self.named_steps[node_name]
            if isinstance(node, BasePipeline):
                # If dealing with a sub pipe,
                # Call the child _check_init_params_honored with the updated config
                node._check_init_params_honored(
                    {key.replace("%s:" % node_name, "", 1): value}
                )
                continue

            if isinstance(node, AutoSklearnComponent):
                node_dict = vars(node)
            elif isinstance(node, AutoSklearnChoice):
                node_dict = vars(node.choice)
            else:
                raise ValueError("Unsupported node type {}".format(type(node)))

            if variable_name not in node_dict or node_dict[variable_name] != value:
                raise ValueError(
                    "Cannot properly set the pair {}->{} via init_params"
                    "".format(key, value)
                )

    def __repr__(self):
        class_name = self.__class__.__name__

        configuration = {}
        self.config._populate_values()
        for hp_name in self.config:
            if self.config[hp_name] is not None:
                configuration[hp_name] = self.config[hp_name]

        configuration_string = "".join(
            [
                "configuration={\n  ",
                ",\n  ".join(
                    [
                        "'%s': %s" % (hp_name, repr(configuration[hp_name]))
                        for hp_name in sorted(configuration)
                    ]
                ),
                "}",
            ]
        )

        if len(self.dataset_properties) > 0:
            dataset_properties_string = []
            dataset_properties_string.append("dataset_properties={")
            for i, item in enumerate(self.dataset_properties.items()):
                if i != 0:
                    dataset_properties_string.append(",\n  ")
                else:
                    dataset_properties_string.append("\n  ")

                if isinstance(item[1], str):
                    dataset_properties_string.append("'%s': '%s'" % (item[0], item[1]))
                else:
                    dataset_properties_string.append("'%s': %s" % (item[0], item[1]))
            dataset_properties_string.append("}")
            dataset_properties_string = "".join(dataset_properties_string)

            return_value = "%s(%s,\n%s)" % (
                class_name,
                configuration,
                dataset_properties_string,
            )
        else:
            return_value = "%s(%s)" % (class_name, configuration_string)

        return return_value

    def _get_pipeline_steps(
        self, dataset_properties, feat_type: Optional[FEAT_TYPE_TYPE] = None
    ):
        raise NotImplementedError()

    def _get_estimator_hyperparameter_name(self):
        raise NotImplementedError()

    def get_additional_run_info(self):
        """Allows retrieving additional run information from the pipeline.

        Can be overridden by subclasses to return additional information to
        the optimization algorithm.
        """
        return self._additional_run_info

    def _validate_include_exclude_params(self):
        if self.include is not None and self.exclude is not None:
            for key in self.include.keys():
                if key in self.exclude.keys():
                    raise ValueError(
                        "Cannot specify include and exclude for same step '{}'.".format(
                            key
                        )
                    )

        supported_steps = {
            step[0]: step[1]
            for step in self.steps
            if isinstance(step[1], AutoSklearnChoice)
        }
        for arg in ["include", "exclude"]:
            argument = getattr(self, arg)
            if not argument:
                continue
            for key in list(argument.keys()):
                if key not in supported_steps:
                    raise ValueError(
                        "The provided key '{}' in the '{}' argument is not valid. The"
                        " only supported keys for this task are {}".format(
                            key, arg, list(supported_steps.keys())
                        )
                    )

                candidate_components = argument[key]
                if not (
                    isinstance(candidate_components, list) and candidate_components
                ):
                    raise ValueError(
                        "The provided value of the key '{}' in the '{}' argument is "
                        "not valid. The value must be a non-empty list.".format(
                            key, arg
                        )
                    )

                available_components = list(
                    supported_steps[key]
                    .get_available_components(
                        dataset_properties=self.dataset_properties
                    )
                    .keys()
                )
                for component in candidate_components:
                    if component not in available_components:
                        raise ValueError(
                            "The provided component '{}' for the key '{}' in the '{}'"
                            " argument is not valid. The supported components for the"
                            " step '{}' for this task are {}".format(
                                component, key, arg, key, available_components
                            )
                        )



import numpy as np
from scipy import sparse

from autosklearn.pipeline.components.data_preprocessing.feature_type_numerical import (
    NumericalPreprocessingPipeline,
)

import unittest


class NumericalPreprocessingPipelineTest(unittest.TestCase):
    def test_data_type_consistency(self):
        X = np.random.rand(3, 4)
        Y = NumericalPreprocessingPipeline(
            feat_type={0: "numerical", 1: "numerical", 2: "numerical"}
        ).fit_transform(X)
        self.assertFalse(sparse.issparse(Y))

        X = sparse.csc_matrix(
            ([3.0, 6.0, 4.0, 5.0], ([0, 1, 2, 1], [3, 2, 1, 0])), shape=(3, 4)
        )
        Y = NumericalPreprocessingPipeline(
            feat_type={0: "numerical", 1: "numerical", 2: "numerical"}
        ).fit_transform(X)
        self.assertTrue(sparse.issparse(Y))

    def test_fit_transform(self):
        X = np.array(
            [[3.14, 1.0, 1.0], [3.14, 2.0, np.nan], [3.14, 3.0, 3.0]]
        )  # noqa : matrix legibility
        # 1st column should be droped due to low variance
        # The 2nd should be be standardized (default rescaling algorithm)
        # The 3rd will get a value imputed by the mean (2.), therefore the
        # transformation here will have the same effect as on the the 2nd column
        sdev = np.sqrt(2 / 3)
        Y1 = np.array(
            [
                [-1 / sdev, -1 / sdev],
                [0.0, 0.0],  # noqa : matrix legibility
                [1 / sdev, 1 / sdev],
            ]
        )  # noqa : matrix legibility
        # dense input
        Yt = NumericalPreprocessingPipeline(
            feat_type={0: "numerical", 1: "numerical", 2: "numerical"}
        ).fit_transform(X)
        np.testing.assert_array_almost_equal(Yt, Y1)
        # sparse input (uses with_mean=False)
        Y2 = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]]) / sdev
        X_sparse = sparse.csc_matrix(X)
        Yt = NumericalPreprocessingPipeline(
            feat_type={0: "numerical", 1: "numerical", 2: "numerical"}
        ).fit_transform(X_sparse)
        np.testing.assert_array_almost_equal(Yt.todense(), Y2)

    def test_transform(self):
        X1 = np.array(
            [[3.14, 1.0, 1.0], [3.14, 2.0, np.nan], [3.14, 3.0, 3.0]]
        )  # noqa : matrix legibility
        sdev = np.sqrt(2 / 3)
        # fit
        NPP = NumericalPreprocessingPipeline(
            feat_type={0: "numerical", 1: "numerical", 2: "numerical"}
        )
        NPP.fit_transform(X1)
        # transform
        X2 = np.array([[1.0, 5.0, 8.0], [2.0, 6.0, 9.0], [3.0, 7.0, np.nan]])
        Yt = NPP.transform(X2)
        # imputation, variance_threshold and rescaling are done using the data already
        # fitted, therefore:
        Y2 = np.array(
            [[3 / sdev, 6 / sdev], [4 / sdev, 7 / sdev], [5 / sdev, 0.0]]
        )  # noqa : matrix legibility
        np.testing.assert_array_almost_equal(Yt, Y2)



from typing import Optional

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformIntegerHyperparameter,
)

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA
from autosklearn.util.common import check_for_bool


class PolynomialFeatures(AutoSklearnPreprocessingAlgorithm):
    def __init__(self, degree, interaction_only, include_bias, random_state=None):
        self.degree = degree
        self.interaction_only = interaction_only
        self.include_bias = include_bias

        self.random_state = random_state
        self.preprocessor = None

    def fit(self, X, Y):
        import sklearn.preprocessing

        self.degree = int(self.degree)
        self.interaction_only = check_for_bool(self.interaction_only)
        self.include_bias = check_for_bool(self.include_bias)

        self.preprocessor = sklearn.preprocessing.PolynomialFeatures(
            degree=self.degree,
            interaction_only=self.interaction_only,
            include_bias=self.include_bias,
        )
        self.preprocessor.fit(X, Y)
        return self

    def transform(self, X):
        if self.preprocessor is None:
            raise NotImplementedError()
        return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "PolynomialFeatures",
            "name": "PolynomialFeatures",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None, dataset_properties=None
    ):
        # More than degree 3 is too expensive!
        degree = UniformIntegerHyperparameter("degree", 2, 3, 2)
        interaction_only = CategoricalHyperparameter(
            "interaction_only", ["False", "True"], "False"
        )
        include_bias = CategoricalHyperparameter(
            "include_bias", ["True", "False"], "True"
        )

        cs = ConfigurationSpace()
        cs.add_hyperparameters([degree, interaction_only, include_bias])

        return cs



"""Test the performance over time functionality of automl instances"""
from autosklearn.automl import AutoML

from pytest_cases import parametrize_with_cases
from pytest_cases.filters import has_tag

import test.test_automl.cases as cases


@parametrize_with_cases(
    "automl",
    cases=cases,
    filter=has_tag("fitted") & ~has_tag("no_ensemble"),
)
def test_performance_over_time_with_ensemble(automl: AutoML) -> None:
    """
    Parameters
    ----------
    automl: AutoMLClassifier
        The fitted automl instance with an ensemble

    Expects
    -------
    * Performance over time should include only the given columns
    * The performance over time should have at least one entry that isn't NaN
    * The timestamps should be monotonic
    """
    expected_performance_columns = {
        "single_best_train_score",
        "single_best_optimization_score",
        "ensemble_optimization_score",
        "Timestamp",
    }
    columns = automl.performance_over_time_.columns
    assert set(columns) == set(expected_performance_columns)

    perf_over_time = automl.performance_over_time_
    assert len(perf_over_time.drop(columns="Timestamp").dropna()) != 0
    assert perf_over_time["Timestamp"].is_monotonic



from __future__ import annotations

from typing import Callable

from functools import partial

from dask.distributed import Client, get_client

from pytest import FixtureRequest, fixture

# Terrible practice but we need to close dask clients somehow
active_clients: dict[str, Callable] = {}


@fixture(autouse=True)
def clean_up_any_dask_clients(request: FixtureRequest) -> None:
    """Auto injected fixture to close dask clients after each test"""
    yield  # Give control to the function

    # Initiate cleanup
    id = request.node.nodeid
    if id in active_clients:
        if request.config.getoption("verbose") > 1:
            print(f"\nFixture closing dask_client for {id}")

        close = active_clients[id]
        close()


def create_test_dask_client(
    id: str,
    n_workers: int = 2,
) -> Client:
    """Factory to make a Dask client and a function to close it
    them.

    Parameters
    ----------
    id: str
        An id to associate with this dask client

    n_workers: int = 2

    Returns
    -------
    Client
        The client
    """
    # Workers are in subprocesses to not create deadlocks with the pynisher
    # and logging.
    client = Client(
        n_workers=n_workers,
        threads_per_worker=1,
        processes=False,
        scheduler_port=0,  # Set to 0 so it chooses a random one
        dashboard_address=None,  # Disable dashboarding
    )
    adr = client.scheduler_info()["address"]

    def close() -> None:
        try:
            client = get_client(adr, timeout=1)
            client.shutdown()
        except Exception:
            pass

    active_clients[id] = close

    return client


@fixture
def make_dask_client(request: FixtureRequest) -> Callable[[int], Client]:
    """Factory to make a Dask client and a function to close it

    Parameters
    ----------
    n_workers: int = 1
        How many workers to have in the dask client

    Returns
    -------
    Client, Callable
        The client and a function to call to close that client
    """
    return partial(create_test_dask_client, id=request.node.nodeid)


# TODO remove in favour of make_dask_client
@fixture(scope="function")
def dask_client(make_dask_client: Callable) -> Client:
    """Create a dask client with two workers."""
    client = make_dask_client(n_workers=2)
    yield client


# TODO remove in favour of make_dask_client
@fixture(scope="function")
def dask_client_single_worker(make_dask_client: Callable) -> Client:
    """Dask client with only 1 worker

    Note
    ----
    May create deadlocks with logging and pynisher
    """
    client = make_dask_client(n_workers=1)
    yield client



import sklearn.ensemble

from autosklearn.pipeline.components.classification.gradient_boosting import (
    GradientBoostingClassifier,
)

from .test_base import BaseClassificationComponentTest


class GradientBoostingComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.92
    res["default_iris_iterative"] = 0.92
    res["default_iris_proba"] = 1.1099521844626845
    res["default_iris_sparse"] = -1
    res["default_digits"] = 0.8652094717668488
    res["default_digits_iterative"] = 0.8652094717668488
    res["default_digits_binary"] = 0.9933211900425015
    res["default_digits_multilabel"] = -1
    res["default_digits_multilabel_proba"] = -1

    sk_mod = sklearn.ensemble.ExtraTreesClassifier
    module = GradientBoostingClassifier
    step_hyperparameter = {
        "name": "max_iter",
        "value": module.get_max_iter(),
    }



from scipy import sparse

from autosklearn.pipeline.components.data_preprocessing.imputation.numerical_imputation import (  # noqa: E501
    NumericalImputation,
)
from autosklearn.pipeline.util import PreprocessingTestCase, _test_preprocessing


class NumericalImputationTest(PreprocessingTestCase):
    def test_default_configuration(self):
        transformations = []
        for i in range(2):
            transformation, original = _test_preprocessing(NumericalImputation)
            self.assertEqual(transformation.shape, original.shape)
            self.assertTrue((transformation == original).all())
            transformations.append(transformation)
            if len(transformations) > 1:
                self.assertTrue((transformations[-1] == transformations[-2]).all())

    def test_default_configuration_sparse_data(self):
        transformations = []
        transformation, original = _test_preprocessing(
            NumericalImputation, make_sparse=True
        )
        self.assertEqual(transformation.shape, original.shape)
        self.assertTrue((transformation.data == original.data).all())
        self.assertIsInstance(transformation, sparse.csc_matrix)
        transformations.append(transformation)

    def test_preprocessing_dtype(self):
        super(NumericalImputationTest, self)._test_preprocessing_dtype(
            NumericalImputation, add_NaNs=True
        )
